/*
 * This file was generated by qdbusxml2cpp version 0.8
 * Source file was org.freedesktop.DBus.Peer.xml
 *
 * qdbusxml2cpp is Copyright (C) The Qt Company Ltd. and other contributors.
 *
 * This is an auto-generated file.
 * This file may have been hand-edited. Look for HAND-EDIT comments
 * before re-generating it.
 */

#include "src/linglong/api/dbus/v1/gen_org_freedesktop_dbus_peer.h"

/*
 * Implementation of interface class OrgFreedesktopDBusPeerInterface
 */

OrgFreedesktopDBusPeerInterface::OrgFreedesktopDBusPeerInterface(const QString &service, const QString &path, const QDBusConnection &connection, QObject *parent)
    : QDBusAbstractInterface(service, path, staticInterfaceName(), connection, parent)
{
}

OrgFreedesktopDBusPeerInterface::~OrgFreedesktopDBusPeerInterface()
{
}


#include "moc_gen_org_freedesktop_dbus_peer.cpp"
