/****************************************************************************
** Meta object code from reading C++ file 'gen_org_deepin_linglong_task1.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.2)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "gen_org_deepin_linglong_task1.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'gen_org_deepin_linglong_task1.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.2. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN31OrgDeepinLinglongTask1InterfaceE_t {};
} // unnamed namespace

template <> constexpr inline auto OrgDeepinLinglongTask1Interface::qt_create_metaobjectdata<qt_meta_tag_ZN31OrgDeepinLinglongTask1InterfaceE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "OrgDeepinLinglongTask1Interface",
        "RequestInteraction",
        "",
        "interactionId",
        "messageID",
        "QVariantMap",
        "additionalMessage",
        "TaskEvent",
        "event",
        "data",
        "TaskFinished",
        "result",
        "Cancel",
        "QDBusPendingReply<>",
        "ReplyInteraction",
        "replies",
        "Start"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'RequestInteraction'
        QtMocHelpers::SignalData<void(const QString &, int, const QVariantMap &)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 3 }, { QMetaType::Int, 4 }, { 0x80000000 | 5, 6 },
        }}),
        // Signal 'TaskEvent'
        QtMocHelpers::SignalData<void(const QString &, const QVariantMap &)>(7, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 8 }, { 0x80000000 | 5, 9 },
        }}),
        // Signal 'TaskFinished'
        QtMocHelpers::SignalData<void(const QVariantMap &)>(10, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 5, 11 },
        }}),
        // Slot 'Cancel'
        QtMocHelpers::SlotData<QDBusPendingReply<>()>(12, 2, QMC::AccessPublic, 0x80000000 | 13),
        // Slot 'ReplyInteraction'
        QtMocHelpers::SlotData<QDBusPendingReply<>(const QString &, const QVariantMap &)>(14, 2, QMC::AccessPublic, 0x80000000 | 13, {{
            { QMetaType::QString, 3 }, { 0x80000000 | 5, 15 },
        }}),
        // Slot 'Start'
        QtMocHelpers::SlotData<QDBusPendingReply<>()>(16, 2, QMC::AccessPublic, 0x80000000 | 13),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<OrgDeepinLinglongTask1Interface, qt_meta_tag_ZN31OrgDeepinLinglongTask1InterfaceE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject OrgDeepinLinglongTask1Interface::staticMetaObject = { {
    QMetaObject::SuperData::link<QDBusAbstractInterface::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN31OrgDeepinLinglongTask1InterfaceE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN31OrgDeepinLinglongTask1InterfaceE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN31OrgDeepinLinglongTask1InterfaceE_t>.metaTypes,
    nullptr
} };

void OrgDeepinLinglongTask1Interface::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<OrgDeepinLinglongTask1Interface *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->RequestInteraction((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<int>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<QVariantMap>>(_a[3]))); break;
        case 1: _t->TaskEvent((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QVariantMap>>(_a[2]))); break;
        case 2: _t->TaskFinished((*reinterpret_cast<std::add_pointer_t<QVariantMap>>(_a[1]))); break;
        case 3: { QDBusPendingReply<> _r = _t->Cancel();
            if (_a[0]) *reinterpret_cast<QDBusPendingReply<>*>(_a[0]) = std::move(_r); }  break;
        case 4: { QDBusPendingReply<> _r = _t->ReplyInteraction((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QVariantMap>>(_a[2])));
            if (_a[0]) *reinterpret_cast<QDBusPendingReply<>*>(_a[0]) = std::move(_r); }  break;
        case 5: { QDBusPendingReply<> _r = _t->Start();
            if (_a[0]) *reinterpret_cast<QDBusPendingReply<>*>(_a[0]) = std::move(_r); }  break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (OrgDeepinLinglongTask1Interface::*)(const QString & , int , const QVariantMap & )>(_a, &OrgDeepinLinglongTask1Interface::RequestInteraction, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (OrgDeepinLinglongTask1Interface::*)(const QString & , const QVariantMap & )>(_a, &OrgDeepinLinglongTask1Interface::TaskEvent, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (OrgDeepinLinglongTask1Interface::*)(const QVariantMap & )>(_a, &OrgDeepinLinglongTask1Interface::TaskFinished, 2))
            return;
    }
}

const QMetaObject *OrgDeepinLinglongTask1Interface::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *OrgDeepinLinglongTask1Interface::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN31OrgDeepinLinglongTask1InterfaceE_t>.strings))
        return static_cast<void*>(this);
    return QDBusAbstractInterface::qt_metacast(_clname);
}

int OrgDeepinLinglongTask1Interface::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QDBusAbstractInterface::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 6)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 6;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 6)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 6;
    }
    return _id;
}

// SIGNAL 0
void OrgDeepinLinglongTask1Interface::RequestInteraction(const QString & _t1, int _t2, const QVariantMap & _t3)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 0, nullptr, _t1, _t2, _t3);
}

// SIGNAL 1
void OrgDeepinLinglongTask1Interface::TaskEvent(const QString & _t1, const QVariantMap & _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 1, nullptr, _t1, _t2);
}

// SIGNAL 2
void OrgDeepinLinglongTask1Interface::TaskFinished(const QVariantMap & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 2, nullptr, _t1);
}
QT_WARNING_POP
