// This file is generated by tools/codegen.sh
// DO NOT EDIT IT.

// clang-format off

//  To parse this JSON data, first install
//
//      json.hpp  https://github.com/nlohmann/json
//
//  Then include this file, and then do
//
//     CommonOptions.hpp data = nlohmann::json::parse(jsonString);

#pragma once

#include <optional>
#include <nlohmann/json.hpp>
#include "linglong/api/types/v1/helper.hpp"

namespace linglong {
namespace api {
namespace types {
namespace v1 {
/**
* this is common option for Install and InstallFromFile
*/

using nlohmann::json;

/**
* this is common option for Install and InstallFromFile
*/
struct CommonOptions {
/**
* force to overwrite
*/
bool force;
/**
* do not automatically remove unused dependencies
*/
std::optional<bool> noAutoPrune;
/**
* skip interaction, such as 'apt install aa -y'
*/
bool skipInteraction;
};
}
}
}
}

// clang-format on
