// This file is generated by tools/codegen.sh
// DO NOT EDIT IT.

// clang-format off

//  To parse this JSON data, first install
//
//      json.hpp  https://github.com/nlohmann/json
//
//  Then include this file, and then do
//
//     ContainerEdits.hpp data = nlohmann::json::parse(jsonString);

#pragma once

#include <optional>
#include <nlohmann/json.hpp>
#include "linglong/cdi/types/helper.hpp"

#include "linglong/cdi/types/DeviceNode.hpp"
#include "linglong/cdi/types/Hook.hpp"
#include "linglong/cdi/types/IntelRdt.hpp"
#include "linglong/cdi/types/Mount.hpp"
#include "linglong/cdi/types/LinuxNetDevice.hpp"

namespace linglong {
namespace cdi {
namespace types {
using nlohmann::json;

struct ContainerEdits {
std::optional<std::vector<int64_t>> additionalGids;
std::optional<std::vector<DeviceNode>> deviceNodes;
std::optional<std::vector<std::string>> env;
std::optional<std::vector<Hook>> hooks;
std::optional<IntelRdt> intelRdt;
std::optional<std::vector<Mount>> mounts;
std::optional<std::vector<LinuxNetDevice>> netDevices;
};
}
}
}

// clang-format on
