/*
 * This file was generated by qdbusxml2cpp version 0.8
 * Source file was org.deepin.linglong.Task1.xml
 *
 * qdbusxml2cpp is Copyright (C) The Qt Company Ltd. and other contributors.
 *
 * This is an auto-generated file.
 * This file may have been hand-edited. Look for HAND-EDIT comments
 * before re-generating it.
 */

#ifndef GEN_ORG_DEEPIN_LINGLONG_TASK1_H
#define GEN_ORG_DEEPIN_LINGLONG_TASK1_H

#include <QtCore/QObject>
#include <QtDBus/QtDBus>
#include "linglong/package_manager/package_task.h"
#include <QtCore/qcontainerfwd.h>

/*
 * Adaptor class for interface org.deepin.linglong.Task1
 */
class OrgDeepinLinglongTask1Adaptor: public QDBusAbstractAdaptor
{
    Q_OBJECT
    Q_CLASSINFO("D-Bus Interface", "org.deepin.linglong.Task1")
    Q_CLASSINFO("D-Bus Introspection", ""
"  <interface name=\"org.deepin.linglong.Task1\">\n"
"    <method name=\"Start\"/>\n"
"    <method name=\"Cancel\"/>\n"
"    <method name=\"ReplyInteraction\">\n"
"      <arg direction=\"in\" type=\"s\" name=\"interactionId\"/>\n"
"      <arg direction=\"in\" type=\"a{sv}\" name=\"replies\"/>\n"
"      <annotation value=\"QVariantMap\" name=\"org.qtproject.QtDBus.QtTypeName.In1\"/>\n"
"    </method>\n"
"    <signal name=\"TaskEvent\">\n"
"      <arg type=\"s\" name=\"event\"/>\n"
"      <arg type=\"a{sv}\" name=\"data\"/>\n"
"      <annotation value=\"QVariantMap\" name=\"org.qtproject.QtDBus.QtTypeName.Out1\"/>\n"
"    </signal>\n"
"    <signal name=\"TaskFinished\">\n"
"      <arg type=\"a{sv}\" name=\"result\"/>\n"
"      <annotation value=\"QVariantMap\" name=\"org.qtproject.QtDBus.QtTypeName.Out0\"/>\n"
"    </signal>\n"
"    <signal name=\"RequestInteraction\">\n"
"      <arg type=\"s\" name=\"interactionId\"/>\n"
"      <arg type=\"i\" name=\"messageID\"/>\n"
"      <arg type=\"a{sv}\" name=\"additionalMessage\"/>\n"
"      <annotation value=\"QVariantMap\" name=\"org.qtproject.QtDBus.QtTypeName.Out2\"/>\n"
"    </signal>\n"
"  </interface>\n"
        "")
public:
    OrgDeepinLinglongTask1Adaptor(linglong::service::PackageTask *parent);
    ~OrgDeepinLinglongTask1Adaptor() override;

    inline linglong::service::PackageTask *parent() const
    { return static_cast<linglong::service::PackageTask *>(QObject::parent()); }

public: // PROPERTIES
public Q_SLOTS: // METHODS
    void Cancel();
    void ReplyInteraction(const QString &interactionId, const QVariantMap &replies);
    void Start();
Q_SIGNALS: // SIGNALS
    void RequestInteraction(const QString &interactionId, int messageID, const QVariantMap &additionalMessage);
    void TaskEvent(const QString &event, const QVariantMap &data);
    void TaskFinished(const QVariantMap &result);
};

#endif
