/****************************************************************************
** Meta object code from reading C++ file 'gen_org_deepin_linglong_task1.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.2)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "gen_org_deepin_linglong_task1.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'gen_org_deepin_linglong_task1.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.2. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN29OrgDeepinLinglongTask1AdaptorE_t {};
} // unnamed namespace

template <> constexpr inline auto OrgDeepinLinglongTask1Adaptor::qt_create_metaobjectdata<qt_meta_tag_ZN29OrgDeepinLinglongTask1AdaptorE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "OrgDeepinLinglongTask1Adaptor",
        "D-Bus Interface",
        "org.deepin.linglong.Task1",
        "D-Bus Introspection",
        "  <interface name=\"org.deepin.linglong.Task1\">\n    <method name"
        "=\"Start\"/>\n    <method name=\"Cancel\"/>\n    <method name=\"Re"
        "plyInteraction\">\n      <arg direction=\"in\" type=\"s\" name=\"i"
        "nteractionId\"/>\n      <arg direction=\"in\" type=\"a{sv}\" name="
        "\"replies\"/>\n      <annotation value=\"QVariantMap\" name=\"org."
        "qtproject.QtDBus.QtTypeName.In1\"/>\n    </method>\n    <signal na"
        "me=\"TaskEvent\">\n      <arg type=\"s\" name=\"event\"/>\n      <"
        "arg type=\"a{sv}\" name=\"data\"/>\n      <annotation value=\"QVar"
        "iantMap\" name=\"org.qtproject.QtDBus.QtTypeName.Out1\"/>\n    </s"
        "ignal>\n    <signal name=\"TaskFinished\">\n      <arg type=\"a{sv"
        "}\" name=\"result\"/>\n      <annotation value=\"QVariantMap\" nam"
        "e=\"org.qtproject.QtDBus.QtTypeName.Out0\"/>\n    </signal>\n    <"
        "signal name=\"RequestInteraction\">\n      <arg type=\"s\" name=\""
        "interactionId\"/>\n      <arg type=\"i\" name=\"messageID\"/>\n   "
        "   <arg type=\"a{sv}\" name=\"additionalMessage\"/>\n      <annota"
        "tion value=\"QVariantMap\" name=\"org.qtproject.QtDBus.QtTypeName."
        "Out2\"/>\n    </signal>\n  </interface>\n",
        "RequestInteraction",
        "",
        "interactionId",
        "messageID",
        "QVariantMap",
        "additionalMessage",
        "TaskEvent",
        "event",
        "data",
        "TaskFinished",
        "result",
        "Cancel",
        "ReplyInteraction",
        "replies",
        "Start"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'RequestInteraction'
        QtMocHelpers::SignalData<void(const QString &, int, const QVariantMap &)>(5, 6, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 7 }, { QMetaType::Int, 8 }, { 0x80000000 | 9, 10 },
        }}),
        // Signal 'TaskEvent'
        QtMocHelpers::SignalData<void(const QString &, const QVariantMap &)>(11, 6, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 12 }, { 0x80000000 | 9, 13 },
        }}),
        // Signal 'TaskFinished'
        QtMocHelpers::SignalData<void(const QVariantMap &)>(14, 6, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 9, 15 },
        }}),
        // Slot 'Cancel'
        QtMocHelpers::SlotData<void()>(16, 6, QMC::AccessPublic, QMetaType::Void),
        // Slot 'ReplyInteraction'
        QtMocHelpers::SlotData<void(const QString &, const QVariantMap &)>(17, 6, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 7 }, { 0x80000000 | 9, 18 },
        }}),
        // Slot 'Start'
        QtMocHelpers::SlotData<void()>(19, 6, QMC::AccessPublic, QMetaType::Void),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    QtMocHelpers::UintData qt_constructors {};
    QtMocHelpers::ClassInfos qt_classinfo({
            {    1,    2 },
            {    3,    4 },
    });
    return QtMocHelpers::metaObjectData<OrgDeepinLinglongTask1Adaptor, qt_meta_tag_ZN29OrgDeepinLinglongTask1AdaptorE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums, qt_constructors, qt_classinfo);
}
Q_CONSTINIT const QMetaObject OrgDeepinLinglongTask1Adaptor::staticMetaObject = { {
    QMetaObject::SuperData::link<QDBusAbstractAdaptor::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN29OrgDeepinLinglongTask1AdaptorE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN29OrgDeepinLinglongTask1AdaptorE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN29OrgDeepinLinglongTask1AdaptorE_t>.metaTypes,
    nullptr
} };

void OrgDeepinLinglongTask1Adaptor::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<OrgDeepinLinglongTask1Adaptor *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->RequestInteraction((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<int>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<QVariantMap>>(_a[3]))); break;
        case 1: _t->TaskEvent((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QVariantMap>>(_a[2]))); break;
        case 2: _t->TaskFinished((*reinterpret_cast<std::add_pointer_t<QVariantMap>>(_a[1]))); break;
        case 3: _t->Cancel(); break;
        case 4: _t->ReplyInteraction((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QVariantMap>>(_a[2]))); break;
        case 5: _t->Start(); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (OrgDeepinLinglongTask1Adaptor::*)(const QString & , int , const QVariantMap & )>(_a, &OrgDeepinLinglongTask1Adaptor::RequestInteraction, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (OrgDeepinLinglongTask1Adaptor::*)(const QString & , const QVariantMap & )>(_a, &OrgDeepinLinglongTask1Adaptor::TaskEvent, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (OrgDeepinLinglongTask1Adaptor::*)(const QVariantMap & )>(_a, &OrgDeepinLinglongTask1Adaptor::TaskFinished, 2))
            return;
    }
}

const QMetaObject *OrgDeepinLinglongTask1Adaptor::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *OrgDeepinLinglongTask1Adaptor::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN29OrgDeepinLinglongTask1AdaptorE_t>.strings))
        return static_cast<void*>(this);
    return QDBusAbstractAdaptor::qt_metacast(_clname);
}

int OrgDeepinLinglongTask1Adaptor::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QDBusAbstractAdaptor::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 6)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 6;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 6)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 6;
    }
    return _id;
}

// SIGNAL 0
void OrgDeepinLinglongTask1Adaptor::RequestInteraction(const QString & _t1, int _t2, const QVariantMap & _t3)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 0, nullptr, _t1, _t2, _t3);
}

// SIGNAL 1
void OrgDeepinLinglongTask1Adaptor::TaskEvent(const QString & _t1, const QVariantMap & _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 1, nullptr, _t1, _t2);
}

// SIGNAL 2
void OrgDeepinLinglongTask1Adaptor::TaskFinished(const QVariantMap & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 2, nullptr, _t1);
}
QT_WARNING_POP
