/*
 * This file was generated by qdbusxml2cpp version 0.8
 * Source file was org.deepin.linglong.Task1.xml
 *
 * qdbusxml2cpp is Copyright (C) The Qt Company Ltd. and other contributors.
 *
 * This is an auto-generated file.
 * Do not edit! All changes made to it will be lost.
 */

#include "src/linglong/adaptors/task/gen_org_deepin_linglong_task1.h"
#include <QtCore/QMetaObject>
#include <QtCore/QByteArray>
#include <QtCore/QList>
#include <QtCore/QMap>
#include <QtCore/QString>
#include <QtCore/QStringList>
#include <QtCore/QVariant>

/*
 * Implementation of adaptor class OrgDeepinLinglongTask1Adaptor
 */

OrgDeepinLinglongTask1Adaptor::OrgDeepinLinglongTask1Adaptor(linglong::service::PackageTask *parent)
    : QDBusAbstractAdaptor(parent)
{
    // constructor
    setAutoRelaySignals(true);
}

OrgDeepinLinglongTask1Adaptor::~OrgDeepinLinglongTask1Adaptor()
{
    // destructor
}

void OrgDeepinLinglongTask1Adaptor::Cancel()
{
    // handle method call org.deepin.linglong.Task1.Cancel
    parent()->Cancel();
}

void OrgDeepinLinglongTask1Adaptor::ReplyInteraction(const QString &interactionId, const QVariantMap &replies)
{
    // handle method call org.deepin.linglong.Task1.ReplyInteraction
    parent()->ReplyInteraction(interactionId, replies);
}

void OrgDeepinLinglongTask1Adaptor::Start()
{
    // handle method call org.deepin.linglong.Task1.Start
    parent()->Start();
}


#include "moc_gen_org_deepin_linglong_task1.cpp"
